@@ -7,7 +7,7 @@ import signal
 signal.signal(signal.SIGINT, signal.SIG_DFL)
 
 import sys, os, datetime, time, json, traceback, csv
-import multiprocessing, subprocess, re, ctypes
+import multiprocessing, subprocess, re, ctypes, numbers
 from optparse import OptionParser
 from ._backup import *
 import rethinkdb as r
@@ -29,7 +29,8 @@ except ImportError:
 info = "'rethinkdb export` exports data from a RethinkDB cluster into a directory"
 usage = "\
   rethinkdb export [-c HOST:PORT] [-a AUTH_KEY] [-d DIR] [-e (DB | DB.TABLE)]...\n\
-      [--format (csv | json)] [--fields FIELD,FIELD...] [--clients NUM]"
+      [--format (csv | json)] [--fields FIELD,FIELD...] [--delimiter CHARACTER]\n\
+      [--clients NUM]"
 
 def print_export_help():
     print(info)
@@ -49,6 +50,9 @@ def print_export_help():
     print("  --clients NUM                    number of tables to export simultaneously (defaults")
     print("                                   to 3)")
     print("")
+    print("Export in CSV format:")
+    print("  --delimiter CHARACTER            character to be used as field delimiter, or '\\t' for tab")
+    print("")
     print("EXAMPLES:")
     print("rethinkdb export -c mnemosyne:39500")
     print("  Export all data from a cluster running on host 'mnemosyne' with a client port at 39500.")
@@ -59,8 +63,9 @@ def print_export_help():
     print("rethinkdb export -c hades -e test.subscribers -a hunter2")
     print("  Export a specific table from a cluster running on host 'hades' which requires authorization.")
     print("")
-    print("rethinkdb export --format csv -e test.history --fields time,message")
-    print("  Export a specific table from a local cluster in CSV format with the fields 'time' and 'message'.")
+    print("rethinkdb export --format csv -e test.history --fields time,message --delimiter ';'")
+    print("  Export a specific table from a local cluster in CSV format with the fields 'time' and 'message',")
+    print("  using a semicolon as field delimiter (rather than a comma).")
     print("")
     print("rethinkdb export --fields id,value -e test.data")
     print("  Export a specific table from a local cluster in JSON format with only the fields 'id' and 'value'.")
@@ -73,6 +78,7 @@ def parse_options():
     parser.add_option("-d", "--directory", dest="directory", metavar="DIRECTORY", default=None, type="string")
     parser.add_option("-e", "--export", dest="tables", metavar="DB | DB.TABLE", default=[], action="append", type="string")
     parser.add_option("--fields", dest="fields", metavar="<FIELD>,<FIELD>...", default=None, type="string")
+    parser.add_option("--delimiter", dest="delimiter", metavar="CHARACTER", default=None, type="string")
     parser.add_option("--clients", dest="clients", metavar="NUM", default=3, type="int")
     parser.add_option("-h", "--help", dest="help", default=False, action="store_true")
     parser.add_option("--debug", dest="debug", default=False, action="store_true")
@@ -122,6 +128,23 @@ def parse_options():
     else:
         res["fields"] = options.fields.split(",")
 
+    if options.delimiter is None:
+        res["delimiter"] = ","
+
+    else:
+
+        if options.format != "csv":
+            raise RuntimeError("Error: --delimiter option is only valid for CSV file formats")
+
+        if len(options.delimiter) == 1:
+            res["delimiter"] = options.delimiter
+
+        elif options.delimiter == "\\t":
+            res["delimiter"] = "\t"
+
+        else:
+            raise RuntimeError("Error: Must specify only one character for the --delimiter option")
+
     # Get number of clients
     if options.clients < 1:
         raise RuntimeError("Error: invalid number of clients (%d), must be greater than zero" % options.clients)
@@ -135,13 +158,15 @@ def parse_options():
 # connection errors occur.  Don't bother setting progress, because this is a
 # fairly small operation.
 def get_tables(progress, conn, tables):
-    dbs = r.db_list().run(conn)
+    dbs = r.db_list().filter(r.row.ne('rethinkdb')).run(conn)
     res = []
 
     if len(tables) == 0:
         tables = [(db, None) for db in dbs]
 
     for db_table in tables:
+        if db_table[0] == 'rethinkdb':
+            raise RuntimeError("Error: Cannot export tables from the system database: 'rethinkdb'")
         if db_table[0] not in dbs:
             raise RuntimeError("Error: Database '%s' not found" % db_table[0])
 
@@ -174,6 +199,10 @@ def finalize_directory(base_path, base_path_partial):
 # succeed or fail, there is no partial success.
 def write_table_metadata(progress, conn, db, table, base_path):
     table_info = r.db(db).table(table).info().run(conn)
+
+    # Rather than just the index names, store all index information
+    table_info['indexes'] = r.db(db).table(table).index_status().run(conn, binary_format='raw')
+
     out = open(base_path + "/%s/%s.info" % (db, table), "w")
     out.write(json.dumps(table_info) + "\n")
     out.close()
@@ -205,17 +234,17 @@ def read_table_into_queue(progress, conn, db, table, pkey, task_queue, progress_
     finally:
         progress_info[0].value += read_rows % 20
 
+    # Export is done - since we used estimates earlier, update the actual table size
+    progress_info[1].value = progress_info[0].value
+
 def json_writer(filename, fields, task_queue, error_queue):
     try:
         with open(filename, "w") as out:
             first = True
             out.write("[")
-            while True:
-                item = task_queue.get()
-                if len(item) != 1:
-                    break
+            item = task_queue.get()
+            while not isinstance(item, StopIteration):
                 row = item[0]
-
                 if fields is not None:
                     for item in list(row.keys()):
                         if item not in fields:
@@ -225,39 +254,50 @@ def json_writer(filename, fields, task_queue, error_queue):
                     out.write("\n" + json.dumps(row))
                 else:
                     out.write(",\n" + json.dumps(row))
+
+                item = task_queue.get()
             out.write("\n]\n")
     except:
         ex_type, ex_class, tb = sys.exc_info()
         error_queue.put((ex_type, ex_class, traceback.extract_tb(tb)))
 
-def csv_writer(filename, fields, task_queue, error_queue):
+        # Read until the exit task so the readers do not hang on pushing onto the queue
+        while not isinstance(task_queue.get(), StopIteration):
+            pass
+
+def csv_writer(filename, fields, delimiter, task_queue, error_queue):
     try:
         with open(filename, "w") as out:
-            out_writer = csv.writer(out)
-            out_writer.writerow([s.encode('utf-8') for s in fields])
+            out_writer = csv.writer(out, delimiter=delimiter)
+            out_writer.writerow(fields)
 
-            while True:
-                item = task_queue.get()
-                if len(item) != 1:
-                    break
+            item = task_queue.get()
+            while not isinstance(item, StopIteration):
                 row = item[0]
                 info = []
                 # If the data is a simple type, just write it directly, otherwise, write it as json
                 for field in fields:
                     if field not in row:
                         info.append(None)
-                    elif isinstance(row[field], (int, long, float, complex)):
-                        info.append(str(row[field]).encode('utf-8'))
-                    elif isinstance(row[field], (str, unicode)):
+                    elif isinstance(row[field], numbers.Number):
+                        info.append(str(row[field]))
+                    elif isinstance(row[field], str):
+                        info.append(row[field])
+                    elif isinstance(row[field], unicode):
                         info.append(row[field].encode('utf-8'))
                     else:
                         info.append(json.dumps(row[field]))
                 out_writer.writerow(info)
+                item = task_queue.get()
     except:
         ex_type, ex_class, tb = sys.exc_info()
         error_queue.put((ex_type, ex_class, traceback.extract_tb(tb)))
 
-def launch_writer(format, directory, db, table, fields, task_queue, error_queue):
+        # Read until the exit task so the readers do not hang on pushing onto the queue
+        while not isinstance(task_queue.get(), StopIteration):
+            pass
+
+def launch_writer(format, directory, db, table, fields, delimiter, task_queue, error_queue):
     if format == "json":
         filename = directory + "/%s/%s.json" % (db, table)
         return multiprocessing.Process(target=json_writer,
@@ -265,16 +305,17 @@ def launch_writer(format, directory, db, table, fields, task_queue, error_queue)
     elif format == "csv":
         filename = directory + "/%s/%s.csv" % (db, table)
         return multiprocessing.Process(target=csv_writer,
-                                       args=(filename, fields, task_queue, error_queue))
+                                       args=(filename, fields, delimiter, task_queue, error_queue))
     else:
         raise RuntimeError("unknown format type: %s" % format)
 
 def get_table_size(progress, conn, db, table, progress_info):
-    table_size = r.db(db).table(table).count().run(conn)
-    progress_info[1].value = table_size
+    table_size = r.db(db).table(table).info()['doc_count_estimates'].sum().run(conn)
+    progress_info[1].value = int(table_size)
     progress_info[0].value = 0
 
-def export_table(host, port, auth_key, db, table, directory, fields, format, error_queue, progress_info, stream_semaphore, exit_event):
+def export_table(host, port, auth_key, db, table, directory, fields, delimiter, format,
+                 error_queue, progress_info, sindex_counter, stream_semaphore, exit_event):
     writer = None
 
     try:
@@ -283,22 +324,23 @@ def export_table(host, port, auth_key, db, table, directory, fields, format, err
         conn_fn = lambda: r.connect(host, port, auth_key=auth_key)
         rdb_call_wrapper(conn_fn, "count", get_table_size, db, table, progress_info)
         table_info = rdb_call_wrapper(conn_fn, "info", write_table_metadata, db, table, directory)
+        sindex_counter.value += len(table_info["indexes"])
 
         with stream_semaphore:
             task_queue = SimpleQueue()
-            writer = launch_writer(format, directory, db, table, fields, task_queue, error_queue)
+            writer = launch_writer(format, directory, db, table, fields, delimiter, task_queue, error_queue)
             writer.start()
 
             rdb_call_wrapper(conn_fn, "table scan", read_table_into_queue, db, table,
                              table_info["primary_key"], task_queue, progress_info, exit_event)
-    except (r.RqlError, r.RqlDriverError) as ex:
+    except (r.ReqlError, r.ReqlDriverError) as ex:
         error_queue.put((RuntimeError, RuntimeError(ex.message), traceback.extract_tb(sys.exc_info()[2])))
     except:
         ex_type, ex_class, tb = sys.exc_info()
         error_queue.put((ex_type, ex_class, traceback.extract_tb(tb)))
     finally:
         if writer is not None and writer.is_alive():
-            task_queue.put(("exit", "event")) # Exit is triggered by sending a message with two objects
+            task_queue.put(StopIteration())
             writer.join()
 
 def abort_export(signum, frame, exit_event, interrupt_event):
@@ -332,6 +374,7 @@ def run_clients(options, db_table_set):
     error_queue = SimpleQueue()
     interrupt_event = multiprocessing.Event()
     stream_semaphore = multiprocessing.BoundedSemaphore(options["clients"])
+    sindex_counter = multiprocessing.Value(ctypes.c_longlong, 0)
 
     signal.signal(signal.SIGINT, lambda a, b: abort_export(a, b, exit_event, interrupt_event))
     errors = [ ]
@@ -349,9 +392,11 @@ def run_clients(options, db_table_set):
                                                            db, table,
                                                            options["directory_partial"],
                                                            options["fields"],
+                                                           options["delimiter"],
                                                            options["format"],
                                                            error_queue,
                                                            progress_info[-1],
+                                                           sindex_counter,
                                                            stream_semaphore,
                                                            exit_event)))
             processes[-1].start()
@@ -371,12 +416,14 @@ def run_clients(options, db_table_set):
             print_progress(1.0)
 
         # Continue past the progress output line and print total rows processed
-        def plural(num, text):
-            return "%d %s%s" % (num, text, "" if num == 1 else "s")
+        def plural(num, text, plural_text):
+            return "%d %s" % (num, text if num == 1 else plural_text)
 
         print("")
-        print("%s exported from %s" % (plural(sum([max(0, info[0].value) for info in progress_info]), "row"),
-                                       plural(len(db_table_set), "table")))
+        print("%s exported from %s, with %s" %
+              (plural(sum([max(0, info[0].value) for info in progress_info]), "row", "rows"),
+               plural(len(db_table_set), "table", "tables"),
+               plural(sindex_counter.value, "secondary index", "secondary indexes")))
     finally:
         signal.signal(signal.SIGINT, signal.SIG_DFL)
 
@@ -401,6 +448,9 @@ def main():
 
     try:
         conn_fn = lambda: r.connect(options["host"], options["port"], auth_key=options["auth_key"])
+        # Make sure this isn't a pre-`reql_admin` cluster - which could result in data loss
+        # if the user has a database named 'rethinkdb'
+        rdb_call_wrapper(conn_fn, "version check", check_minimum_version, (1, 16, 0))
         db_table_set = rdb_call_wrapper(conn_fn, "table list", get_tables, options["db_tables"])
         del options["db_tables"] # This is not needed anymore, db_table_set is more useful
 
